!
! Copyright (C) 1991-2003  ; All Rights Reserved ; ATMET, LLC
!
! This file is free software; you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation; either version 2 of the License, or (at your option) any later version.
!
! This software is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
! PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with this
! program; if not, write to the Free Software Foundation, Inc.,
! 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
!======================================================================================

program grib_to_gdf

use grib_get_mod
use hdf5_utils

implicit none

integer, parameter :: max_hours=10
character(len=10) :: c_date, c_hr(max_hours)
character(len=128) :: filein,fileout,namein,out_prefix,out_ext,dstring
character(len=2) :: cflag(10)
integer, allocatable :: itlevs(:)
real, allocatable :: a3(:,:,:,:),a2(:,:,:)

integer :: ia,iargc,istr,iend,inc
integer :: i,j,idprep=0,ndate,nhr,ii,nh,k
integer :: nlev,inproj,ivertcoord,miss,ircode,num_hours

integer :: iyyyy,imm,idd,ihh,nuyyyy,numm,nudd,nuhh
real :: tinc, amiss=-999.,xnelat,xnelon

integer, parameter :: maxlev=100, maxvar=100
integer :: iplevs(maxlev), nvar3d, nvar2d, gdf_file_ver, ndims,idims(3)
character(len=4) :: out_format
character(len=10) :: var3d(maxvar), var2d(maxvar)  &
                  , outvar3d(maxvar), outvar2d(maxvar)

namelist/dgrib_in/nvar3d,var3d,nvar2d,var2d,c_date,num_hours,c_hr  &
                 ,filein,out_format,outvar3d,outvar2d  &
                 ,wgrib_exe,out_prefix,out_ext

!!call opngks

ia = iargc  ( )

print*,'num args:',ia

!ugetarg is a RAMS utility that is getarg except for HP when it
!   adds 1 to each argument's index since HP's start from 0, not 1.
!   if compiling without the ugetarg in the RAMS library, use
!   getarg and execute the do loop from 0,ia-1,2.

namein='./DGRIB_IN'
do i=1,ia,2
   call ugetarg(i,cflag(i))
   if(cflag(i)(1:2).eq.'-f')then
      call ugetarg(i+1,namein)
   endif
enddo

c_date='99999999'
num_hours = 1
c_hr(1:max_hours)  ='99999999'

! Set out_format default. regular (old) GDF or HDF5
out_format='GDF'


open(10,file=namein,status='old')
read(10,dgrib_in)
close(10)


do i=1,ia,2
   print*,'args:',i,cflag(i),cflag(i+1)
   call ugetarg(i,cflag(i))
   if(cflag(i)(1:2).eq.'-t')then
      call ugetarg(i+1,c_date)
   elseif(cflag(i)(1:2).eq.'-h')then
      call ugetarg(i+1,c_hr(1))
      num_hours = 1             ! Only 1 time allowed if argument
   elseif(cflag(i)(1:2).eq.'-g')then
      call ugetarg(i+1,filein)
   elseif(cflag(i)(1:2).eq.'-e')then
      call ugetarg(i+1,wgrib_exe)
   elseif(cflag(i)(1:2).eq.'-o')then
      call ugetarg(i+1,out_prefix)
   elseif(cflag(i)(1:2).eq.'-f')then
      ! already processed above
   else
      print*
      print*,'Usage:dgrib [-f namelist_file] [-g gribfile] [-t yyyymmddhh] [-h hh]'
      print*,'      namelist file name defaults to DGRIB_IN'
      print*,'      -t : data time to process - yyyymmddhh'
      print*,'      -h : forecast hour - hh'
      print*,'      -e : wgrib_executable - if ends in "2" assume grib v2'
      print*,'      -o : output file prefix'
      print*,'           If c_date or c_hr set to ''99999999'', '
      print*,'           the date/hour of the first record'
      print*,'           in the grib file will be used.'
      print*
      print*,'      Command line arguments will override the namelist!'
      stop 'fdgrib usage'
   endif
enddo

print*
print*,'---------------------------------------------------------------'
print*,'GRIB to GDF converter'
print*,'---------------------------------------------------------------'
print*,'Namelist file name: ',trim(namein)
print*,'GRIB file name    : ',trim(filein)
print*,'WGRIB exec        : '
if (wgrib_exe(len_trim(wgrib_exe):len_trim(wgrib_exe)) == '2') then
   print*,'      Assumimg GRIB 2   : ',trim(wgrib_exe)
   grib_ver=2
else
   print*,'      Assumimg GRIB 1   : ',trim(wgrib_exe)
   grib_ver=1
endif
print*,'Data date/time     : ',c_date
print*,'Forecast hour      : ',c_hr(1:num_hours)
print*,'3D variables       : ',var3d(1:nvar3d)
print*,'2D variables       : ',var2d(1:nvar2d)
print*,'Output file prefix : ',out_prefix
print*
print*


call grib_get_date(trim(filein))
print*,'dd1'
call grib_get_proj(trim(filein))
print*,'dd2'

call grib_get_fields(trim(filein))

print*
print '(a,a,i4)','Found projection: ',trim(projection), iscan
print '(a,2f12.2,2i5,2f12.4)' &
                  ,'  params:   lat,   lon,  nx,  ny,   dx,   dy'
print '(a,2f12.2,2i5,2f12.4)','        ',alat1,alon1, nx,ny,dx,dy
print*

! Some bookeeping

xnelat=0.
xnelon=0.
ivertcoord=1
if (projection(1:6)=="latlon") then
   inproj=1
   if (iscan == 0 ) alat1=alat1 - dy * (ny-1)
elseif (projection(1:7)=="Lambert") then
   inproj=2
   dx=dx*1000.
   dy=dy*1000.
   aorient=alov
elseif (projection(1:5)=="polar") then
   inproj=3
   ! Assume the polar stereo is true polar stereo
   reflat1=90.
   reflat2=amiss
endif

ndate=0

hour_loop: do nh = 1, num_hours

   nhr=0

   print*
   print*,'Processing date/time:',c_date,c_hr(nh)
   print*

   read(c_hr(nh),'(i10)') nhr
   if(c_hr(nh)   == '99999999') nhr=ifhrs(1)
   if(c_date == '99999999') then
      ndate=longdate
      write(c_date,'(i10)')ndate
   endif

   print*,'dates1:',c_hr(nh),c_date,nhr,ndate
   print*,'dates1:',c_date
   print*,'dates1:',longdate

   !     wgrib only prints out the 2 digit year in short inventory
   !     wgrib2 has 4
   if (grib_ver == 1) read(c_date(3:),'(i8)')ndate
   if (grib_ver == 2) read(c_date(1:),*)ndate

   !     Quick check to see if there are any fields matching time wanted
   print*,'dates2:',nrec
   j=0
   do i=1,nrec
      print*,'recs:',i,trim(fields(i)),idates(i),ndate,ifhrs(i),nhr
      if( trim(fields(i)) == trim(var3d(1)) .and.  &
               idates(i) == ndate .and. ifhrs(i) == nhr)then
         j=1
         exit
      endif
   enddo
   if ( j == 0) then
      print*,'Nothing found of the proper date'
      stop 'no stuff'
   endif

   !     Detmine the number of levels we are going to use
   nlev=0
   do i=1,nrec
      if(levs(i)==0)cycle  !don't count the surface fields here
      if(trim(fields(i)) == trim(var3d(1)) .and.  &
              idates(i) == ndate .and. ifhrs(i) == nhr)then
         nlev=nlev+1
         iplevs(nlev)=levs(i)
      endif
   enddo

   !     Make sure pressure level data monotonically decreases
   allocate (itlevs(nlev))
   itlevs=-1
   do i=1,nlev
      do j=1,nlev
         if(iplevs(j) > itlevs(i))then
            itlevs(i)=iplevs(j)
            ia=j
         endif
      enddo
      !  Effectively remove current highest pressure level
      iplevs(ia)=-1
   enddo
   iplevs(1:nlev)=itlevs(1:nlev)
   deallocate (itlevs)

   print*
   print*,'Pressure levels found:',nlev,iplevs(1:nlev)
   print*

   read(c_date(1:4) ,'(i4)')iyyyy
   read(c_date(5:6) ,'(i2)')imm
   read(c_date(7:8) ,'(i2)')idd
   read(c_date(9:10),'(i2)')ihh

   tinc=real(nhr)
   call date_add_to(iyyyy,imm,idd,ihh*10000,tinc,funit,nuyyyy,numm,nudd,nuhh)
   nuhh=nuhh/100

   allocate (a3(nx,ny,nlev,nvar3d+nvar2d))
   allocate (a2(nx,ny,nvar2d))

   ! Fill the 3 and 2-d variable arrays

   do i=1,nvar3d
      do j=1,nlev
         !print*,'nnn:',var3d(i),nx,ny,j,i,a3(1,1,j,i)
         a3(1:nx,1:ny,j,i)=amiss
         call getfield(trim(filein)  &
                      ,nx*ny,a3(1,1,j,i),var3d(i),iplevs(j),ndate,nhr,ircode)
         !print*,'nnn:',var3d(i),nx,ny,j,i,a3(1,1,j,i)
      enddo
   enddo
   !---------------------------------
   ! The Iberinco ECMWF exception
   if (nvar3d == 5 .and. var3d(6) == 'Z') then
      do j=1,nlev
         ! a3(1:nx,1:ny,j,i)=amiss ! Don't fill in missing data
         call getfield(trim(filein)  &
                      ,nx*ny,a3(1,1,j,4),var3d(6),iplevs(j),ndate,nhr,ircode)
         if (ircode == 1) then
            ! Convert Z to GH
            a3(1:nx,1:ny,j,4) = a3(1:nx,1:ny,j,4)/9.8
         else
            !a3(1:nx,1:ny,j,4)=amiss
         endif
         print*,'zzz:',var3d(6),ircode,j,a3(1,1,j,4)
      enddo
   endif
   !---------------------------------
      !do k=20,20
      !   do j = 1,ny
      !   print*,-34,ny-j+1,j,iplevs(k)  &
      !      ,sqrt(a3(34,j,k,1)**2+a3(34,j,k,2)**2)
      !enddo
      !enddo

   do i=1,nvar2d
      a2(1:nx,1:ny,i)=amiss
      j=0 !specify j=0 for the surface
      call getfield(filein(1:len_trim(filein))  &
                   ,nx*ny,a2(1,1,i),var2d(i),j,ndate,nhr,ircode)
   enddo

   ! Run the arrays through prepfield for possible conversion and flipping
   do i=1,nvar3d
      do j=1,nlev
         call prepfield(a3(1,1,j,i),nx,ny,projection,var3d(i),iscan,amiss)
      enddo
   enddo
      !do k=20,20
      !   do j = 1,ny
      !   print*,-34,ny-j+1,j,iplevs(k)  &
      !      ,sqrt(a3(34,j,k,1)**2+a3(34,j,k,2)**2)
      !enddo
      !enddo

   do i=1,nvar2d
      call prepfield(a2(1,1,i),nx,ny,projection,var2d(i),iscan,amiss)
   enddo

   !-----------------------------------------------------
   ! Output file name
   !-----------------------------------------------------

   write(dstring,45 ) nuyyyy,'-',numm,'-',nudd,'-',nuhh
   45   format(i4.4,a1,i2.2,a1,i2.2,a1,i4.4)
   fileout = trim(out_prefix)//trim(dstring)//trim(out_ext)


   if (out_format == 'GDF') then
      ! Write GDF (vfm)
      print*,'Writing GDF file:',trim(fileout)

      open(10,file=fileout,status='unknown',form='formatted')
      rewind 10

      gdf_file_ver=2
      write(10,*)999999,gdf_file_ver
      write(10,900)iyyyy,imm,idd,ihh*100,nhr*100,nlev,nx,ny &
        ,inproj,dx,dx,alat1,alon1,xnelat,xnelon,reflat1,aorient &
        ,reflat2,ivertcoord,(iplevs(i),i=1,nlev)
      900  format(8i6,/,i1,2x,9(f10.4,2x),/,i1,(12i6))

      !     Go through each variable at each level for writing.

      do j=1,nlev
         do i=1,nvar3d
            write(10,'(8f10.3)') a3(1:nx,1:ny,j,i)
         enddo
      enddo

      do i=1,nvar2d
         write(10,'(8f10.2)') a2(1:nx,1:ny,i)
      enddo

      close(10)

   else
      ! Write HDF5
      print*,'Writing HDF5 file:',fileout

      call shdf5_open(fileout,'W',1)

      gdf_file_ver=3
      ndims=1 ; idims(1)=1
      call shdf5_orec(ndims,idims,'version',ivars=gdf_file_ver)
      call shdf5_orec(ndims,idims,'year'   ,ivars=iyyyy)
      call shdf5_orec(ndims,idims,'month'  ,ivars=imm)
      call shdf5_orec(ndims,idims,'day'    ,ivars=idd)
      call shdf5_orec(ndims,idims,'hour'   ,ivars=ihh*100)
      call shdf5_orec(ndims,idims,'ftime'  ,ivars=nhr*100)
      call shdf5_orec(ndims,idims,'nx'     ,ivars=nx)
      call shdf5_orec(ndims,idims,'ny'     ,ivars=ny)
      call shdf5_orec(ndims,idims,'nlev'   ,ivars=nlev)
      call shdf5_orec(ndims,idims,'iproj'  ,ivars=inproj)
      call shdf5_orec(ndims,idims,'vcoord' ,ivars=ivertcoord)
      call shdf5_orec(ndims,idims,'swlat'  ,rvars=alat1)
      call shdf5_orec(ndims,idims,'swlon'  ,rvars=alon1)
      call shdf5_orec(ndims,idims,'nelat'  ,rvars=xnelat)
      call shdf5_orec(ndims,idims,'nelon'  ,rvars=xnelon)
      call shdf5_orec(ndims,idims,'dx'     ,rvars=dx)
      call shdf5_orec(ndims,idims,'dy'     ,rvars=dx)
      call shdf5_orec(ndims,idims,'reflat1',rvars=reflat1)
      call shdf5_orec(ndims,idims,'reflat2',rvars=reflat2)
      ndims=1 ; idims(1)=nlev
      call shdf5_orec(ndims,idims,'levels'  ,ivara=iplevs)


      do i=1,nvar3d
         ndims=3 ; idims(1)=nx; idims(2)=ny; idims(3)=nlev
         call shdf5_orec(ndims,idims,trim(outvar3d(i)),rvara=a3(1,1,1,i))
      enddo

      do i=1,nvar2d
         ndims=2 ; idims(1)=nx; idims(2)=ny
         print*,'++:',i,trim(outvar2d(i))
         call shdf5_orec(ndims,idims,trim(outvar2d(i)),rvara=a2(1,1,i))
      enddo

      call shdf5_close()

   endif

   !     Write the tag file to indicate file is complete
   open(10,file=trim(out_prefix)//trim(dstring)//'.tag',status='unknown' &
    ,form='formatted')
   write(10,'(a)') 'done'
   close(10)

enddo hour_loop

!!call clsgks
end

!************************************************************************

subroutine getfield(filein,nxy,a,var,iplev,ndate,nhr,ircode)

use grib_get_mod

implicit none

integer :: iplev,ndate,nhr,nxy,ircode
real, dimension(nxy) :: a
character(len=*) :: filein,var

integer :: i

ircode = 0

do i=1,nrec

   if(var == fields(i) .and. iplev == levs(i).and.  &
      ndate == idates(i) .and. nhr == ifhrs(i))then
      print*,'call grib_get_rec:',trim(filein),nxy,a(1),trim(irecs(i))
      call grib_get_recf(trim(filein),trim(irecs(i)),a,nxy)

      print '(a,a,i8,i14,i8)','Found record: ' &
            ,trim(fields(i)),levs(i),idates(i),ifhrs(i)

      ircode = 1
      exit
   endif

enddo

return
end

!************************************************************************

subroutine prepfield(a,nx,ny,projection,type,iscan,amiss)
implicit none

integer nx,ny
real, dimension(nx*ny) :: a
character(len=*) :: type, projection

real, allocatable :: b(:)
real :: amiss
real :: g=9.80616
integer :: iscan

integer i

if(projection(1:6) == 'latlon' .and. iscan == 0)then
   print*,'   flipping the projection'
   allocate (b(nx*ny))
   call flip(nx,ny,a,b)
   a=b
   deallocate (b)
endif

do i=1,nx*ny
   if(a(i) >= 1.e10)a(i)=amiss
enddo

if(type(1:1)=='R')then
   ! Assuming RH in %
   print*,'   factoring R'
   do i=1,nx*ny
      if(a(i) > amiss)a(i)=a(i)/100.
   enddo
elseif(type(1:4)=='SPFH')then
   print*,'   converting spec hum to g/kg'
   do i=1,nx*ny
      if(a(i) > amiss)a(i)=a(i)*1000.
   enddo
elseif(type=='PRMSL'.or.type=='PRES')then
   print*,'   factoring to mb'
   do i=1,nx*ny
      if(a(i) > amiss)a(i)=a(i)/100.
   enddo
elseif(type=='Z')then
   print*,'   factoring geopotential to height using g=',g
   do i=1,nx*ny
      if(a(i) > amiss)a(i)=a(i)/g
   enddo

!!!!!!!!       TO DO:      !!!!!!!!
!
!  If Z is used and geo height less than about -100, this gets
!     confused with the amiss. A suggested change:
!
! if(a(i) > amiss*g) a(i)=a(i)/g
!
!  I don't like operating on missing values. The other option is to change
!     amiss, especially now that hdf5 files are used.


endif

!!call ezcntr(a,nx,ny)


return
end

!************************************************************************

subroutine flip(nx,ny,vin,vout)
implicit none
integer nx,ny
real, dimension(nx,ny) :: vin, vout
integer i,j,ii,jj

do i=1,nx
!  ii=nx/2+i
!  if(ii.gt.nx)ii=i-nx/2
   ii=i
  do j=1,ny
    jj=ny-j+1
    vout(i,j)=vin(ii,jj)
  enddo
enddo
return
end

!************************************************************************

subroutine date_add_to (inyear,inmonth,indate,inhour  &
                        ,tinc,tunits,outyear,outmonth,outdate,outhour)

! add a time increment to a date and output new date
! -> uses hhmmss for hours, 4 digit year

integer inyear,inmonth,indate,inhour  &
       ,outyear,outmonth,outdate,outhour
real tinc
character(len=1) :: tunits
dimension mondays(12)
data mondays/31,28,31,30,31,30,31,31,30,31,30,31/

! convert input time to seconds

ttinc=tinc
if(tunits.eq.'m') ttinc=tinc*60.
if(tunits.eq.'h') ttinc=tinc*3600.
if(tunits.eq.'d') ttinc=tinc*86400.
!print*,'inc:',tinc,tunits,ttinc

xhourin=inhour/10000
xminin=mod(inhour,10000)/100
xsecin=mod(inhour,100)
strtim=xhourin+xminin/60.+xsecin/3600.
!print*,'strtim=',strtim

izhours=int(mod(strtim+ttinc/3600.,24.)+.001)
izmin  =int(mod(strtim+ttinc/3600.,1.)*60+.001)
izsec  =int(mod(strtim*3600.+ttinc,60.)+.001)
!print*,'izs=',izhours,izmin,izsec

outhour= izhours*10000+izmin*100+izsec

iround=.001
if(ttinc<0.) iround=-.001
iadddays=int((strtim+ttinc/3600.)/24.+iround)

outyear=inyear
outdate=indate+iadddays
outmonth=inmonth

20 continue
   idays=mondays(outmonth)
   if(outmonth==2.and.mod(outyear,4)==0)  idays=29

   if(outdate.gt.idays) then
      outdate=outdate-idays
      outmonth=outmonth+1
      if(outmonth.gt.12) then
         outyear=outyear+1
         outmonth=1
      endif
   elseif(outdate.lt.1) then
      if(outmonth.eq.1)outmonth=13
      idays=mondays(outmonth-1)
      if(outmonth-1.eq.2.and.mod(outyear,4).eq.0)  idays=29
      outdate=idays+outdate
      outmonth=outmonth-1
      if(outmonth.eq.12)outyear=outyear-1
   else
      goto 21
   endif

   goto 20

21 continue

!print*,'out stuff:',outyear,outmonth,outdate,outhour

return
end
